import requests
import uuid
import inspect
from typing import Any, Callable, Tuple

from abstra_cli.credentials import get_credentials

def usage(root_path: str) -> Callable[..., Any]:
    def usage_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(*args: Tuple[Any], **kwargs: Any) -> Any:
            arg_names = inspect.getfullargspec(func).args
            arg_values = dict(zip(arg_names, args))

            metric_data = {
            'userId': str(uuid.getnode()),
            'payload': {**arg_values, **kwargs, **{'event': func.__name__}}
            }

            headers = {
                'apiKey': get_credentials(root_path)
            }

            api_url = 'http://cloud-api.abstra.cloud/cli/editor/usage'
            requests.post(api_url, json=metric_data, headers=headers)

            

            return func(*args, **kwargs)
        
        original_route_name = func.__name__
        wrapper.__name__ = f"{original_route_name}_usage_wrapper" ## avoid name collision in blueprints
        
        return wrapper
    return usage_decorator




from ..widget_base import Input
from typing import Optional, Union


class ToggleInput(Input):
    type = "toggle-input"

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.label = label
        self.on_text = kwargs.get("on_text", "Yes")
        self.off_text = kwargs.get("off_text", "No")
        self.initial_value = kwargs.get("initial_value", "")
        self.required = kwargs.get("required", True)
        self.hint = kwargs.get("hint", None)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.disabled = kwargs.get("disabled", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "onText": self.on_text,
            "offText": self.off_text,
            "initialValue": ToggleInput.__revert_value(self.initial_value),
            "required": self.required,
            "hint": self.hint,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
        }

    @staticmethod
    def __revert_value(value: Optional[Union[bool, str]]) -> str:
        if value == "1" or value == True:
            return "1"
        return "0"

    @staticmethod
    def __convert_answer(answer: str) -> str:
        return True if answer == "1" else False

    def convert_answer(self, answer: str) -> str:
        return self.__convert_answer(answer)



import os
from .utils import CREDENTIALS_FILE


def get_credentials(root_path):
    if os.getenv("ABSTRA_API_TOKEN"):
        return os.getenv("ABSTRA_API_TOKEN")

    credentials_path = os.path.join(root_path, CREDENTIALS_FILE)
    if not os.path.exists(credentials_path):
        return None

    with open(credentials_path, encoding="utf-8") as f:
        return f.read().strip()


def delete_credentials(root_path):
    credentials_path = os.path.join(root_path, CREDENTIALS_FILE)
    if os.path.exists(credentials_path):
        os.remove(credentials_path)


def set_credentials(root_path, token):
    credentials_path = os.path.join(root_path, CREDENTIALS_FILE)
    os.makedirs(os.path.dirname(credentials_path), exist_ok=True)

    with open(credentials_path, "w", encoding="utf-8") as f:
        f.write(token)



import unittest
from uuid import uuid4 as uuid
from abstra_server.api import API
from .fixtures import init_dir

class TestForms(unittest.TestCase):
    def updates_across_reloads(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)
        api = API(root=workspace_root_path)

        form = api.create_form()

        api.update_form(form.path, dict(
            title="New Title"
        ))

        api2 = API(root=workspace_root_path)
        api2.get_workspace()
        new_form = api2.get_form(form.path)

        self.assertEqual(form.title, new_form.title)
    
    def test_raise_exception_on_invalid_propery_update(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)
        api = API(root=workspace_root_path)

        form = api.create_form()

        with self.assertRaises(Exception):
            api.update_form(form.path, dict(
                title="New Title",
                invalid_property="invalid"
            ))


import unittest
from .fixtures import init_dir
from uuid import uuid4 as uuid
from abstra_server.api import API

class TestSidebar(unittest.TestCase):
    def test_starts_empty(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)
        api = API(root=workspace_root_path)

        workspace = api.get_workspace()
        self.assertEqual(workspace.__dict__['sidebar'], [])
    
    def test_auto_add_forms(self):
        # given
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)
        api = API(root=workspace_root_path)

        # when
        form = api.create_form()
        workspace = api.get_workspace()

        # then
        sidebar_json = workspace.__dict__['sidebar']
        self.assertEqual(len(sidebar_json), 1)
        item = dict(
            name=form.title,
            path=form.path,
            type='form',
            visible=False
        )
        self.assertEqual(sidebar_json[0] | item, sidebar_json[0])
    
    def test_auto_add_dashes(self):
        # given
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)
        api = API(root=workspace_root_path)

        # when
        dash = api.create_dash()
        workspace = api.get_workspace()

        # then
        sidebar_json = workspace.__dict__['sidebar']
        self.assertEqual(len(sidebar_json), 1)
        item = dict(
            name=dash.title,
            path=dash.path,
            type='dash',
            visible=False
        )
        self.assertEqual(sidebar_json[0] | item, sidebar_json[0])


import json, unittest
from uuid import uuid4 as uuid
from abstra.tables import get_db, get_table, TableNotFound
from .fixtures import init_dir


class TestTables(unittest.TestCase):
    def test_create_table(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        db = get_db()

        created_table = db.create_table()

        retrieved_table = db.get_table(created_table.name)

        self.assertEqual(created_table.name, retrieved_table.name)

    def test_delete_table(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        db = get_db()

        created_table = db.create_table()

        db.delete_table(created_table.name)

        with self.assertRaises(TableNotFound):
            db.get_table(created_table.name)

    def test_update_table_name(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        db = get_db()

        created_table = db.create_table()

        db.update_table(created_table.name, dict(name="new_name"))

        db.get_table("new_name")

    def test_create_column(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        db = get_db()

        table = db.create_table()
        col = db.create_column(table.name)
        cols = db.get_columns(table.name)

        self.assertIn(col.name, [c.name for c in cols])

        db.delete_column(table.name, col.name)

        cols = db.get_columns(table.name)
        self.assertNotIn(col.name, [c.name for c in cols])

    def test_crud_data(self):
        workspace_root_path = f"/tmp/{uuid()}"
        init_dir(workspace_root_path)

        db = get_db()

        table = db.create_table()

        name_column = db.create_column(table.name)
        db.update_column(table.name, name_column.name, dict(name="name"))

        email_column = db.create_column(table.name)
        db.update_column(table.name, email_column.name, dict(name="email"))

        crud_table = get_table(table.name)
        crud_table.insert(
            dict(name="Michael Scott", email="michael.scott@dundermifflin.com")
        )
        crud_table.insert(
            dict(name="Dwight Schrute", email="dwight.schrute@dundermifflin.com")
        )
        crud_table.insert(
            dict(name="Jim Halpert", email="jim.halpert@dundermifflin.com")
        )
        crud_table.insert(dict(name="Pam Beesly", email="pam.beesly@dundermifflin.com"))

        crud_table.insert(dict(name="Uzumaki Naruto", email="uzumaki.naruto@konoha.jp"))
        crud_table.insert(dict(name="Uchiha Sasuke", email="uchiha.sasuke@konoha.jp"))
        crud_table.insert(dict(name="Haruno Sakura", email="haruno.sakura@konoha.jp"))
        crud_table.insert(dict(name="Hatake Kakashi", email="hatake.kakashi@konoha.jp"))

        self.assertEqual(len(crud_table.select()), 8)
        self.assertEqual(
            len(
                crud_table.select(
                    where="email LIKE :email_like",
                    params=dict(email_like="%@dundermifflin.com"),
                )
            ),
            4,
        )
        self.assertEqual(
            crud_table.select(
                columns=["name"],
                where="email = :email",
                params=dict(email="uzumaki.naruto@konoha.jp"),
            ),
            [dict(name="Uzumaki Naruto")],
        )

        crud_table.update(
            where="email = :email",
            params=dict(email="uchiha.sasuke@konoha.jp"),
            set=dict(email="uchiha.sasuke@akatsuki.jp"),
        )

        self.assertEqual(
            crud_table.select(
                columns=["email"],
                where="name = :name",
                params=dict(name="Uchiha Sasuke"),
            ),
            [dict(email="uchiha.sasuke@akatsuki.jp")],
        )

        crud_table.delete(
            where="email = :email", params=dict(email="uchiha.sasuke@akatsuki.jp")
        )

        self.assertEqual(len(crud_table.select()), 7)
